#  Copyright (c) 2013-2019 Canonical Ltd.
#
#  Author: Scott Moser <smoser@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from gettext import gettext as _

from softwareproperties.cloudarchive import CloudArchiveShortcutHandler
from softwareproperties.ppa import PPAShortcutHandler
from softwareproperties.shortcuthandler import InvalidShortcutException
from softwareproperties.sourceslist import SourcesListShortcutHandler
from softwareproperties.uri import URIShortcutHandler


SHORTCUT_HANDLERS = [
    PPAShortcutHandler,
    CloudArchiveShortcutHandler,
    SourcesListShortcutHandler,
    URIShortcutHandler,
]


def shortcut_handler(shortcut, **kwargs):
    for handler in SHORTCUT_HANDLERS:
        try:
            return handler(shortcut, **kwargs)
        except InvalidShortcutException:
            pass

    raise InvalidShortcutException(_("Unable to handle repository shortcut '%s'") % shortcut)


# vi: ts=4 expandtab
